﻿// Author: Orlys
// Github: https://github.com/Orlys

namespace Yuuna.Contracts.Optimization
{
    using System;
    using System.Collections.Generic;
    using System.Collections.Immutable;
    using System.Diagnostics.CodeAnalysis;
    using System.Runtime.CompilerServices;

    using Yuuna.Contracts.Patterns;
    using Yuuna.Contracts.Plugins;
    using Yuuna.Contracts.Semantics;

    public class DefaultStrategy : IStrategy
    {
        public DefaultStrategy()
        {

        }

        public Match? FindBest(IEnumerable<PluginBase> plugins, IImmutableList<string> feed)
        {
            var list = ImmutableArray.CreateBuilder<Match>();
            foreach (var plugin in plugins)
            {
                foreach (var p in plugin.Patterns.ToImmutable())
                {
                    var score = this.CalculatePatternMatch(p, feed);
                    if (score.HasValue)
                        list.Add(score.Value);
                }
            }

            var e = this.SelectBestMatch(list.ToImmutable());
            return e;
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="matches"></param>
        /// <returns></returns>
        protected virtual Match? SelectBestMatch(IImmutableList<Match> matches)
        {
            var sta = new SortedList<Match, int>();
            foreach (var j in matches)
            {
                if (!sta.ContainsKey(j))
                    sta.Add(j, 1);
                else
                    sta[j]++;
            }

            //var tuple = new List<Tuple<Match, double>>();
            //foreach (var s in sta)
            //{
            //    var x = (s.Value) / (double)-(matches.Count * s.Key.SequentialMatches.Count);
            //    var y = Math.Tanh(Math.Pow(Math.E, x));
            //    tuple.Add(Tuple.Create(s.Key, y));
            //}

            //var bestSelect = tuple.OrderByDescending(x => x.Item2).FirstOrDefault();

            //return bestSelect?.Item1;

            var hold = double.MinValue;
            var bestSelect = default(Match?);
            foreach (var s in sta)
            {
                var y = Math.Tanh(Math.Pow(Math.E, (s.Value) / (double)-(matches.Count * s.Key.SequentialMatches.Count)));
                if (y >= hold)
                {
                    hold = y;
                    bestSelect = s.Key;
                }
            }
            return bestSelect;
        }

        /// <summary>
        /// 評估 <paramref name="pattern"/> 與 <paramref name="feed"/> 的符合程度並給予一個分數，並返回一個 <see cref="Match"/> 物件。
        /// 若回傳 <see langword="null"/> 則此規則將不列入計算。
        /// </summary>
        /// <param name="pattern"></param>
        /// <param name="feed"></param>
        /// <returns></returns>
        protected virtual Match? CalculatePatternMatch(IPattern pattern, IImmutableList<string> feed)
        {
            var matches = ImmutableArray.CreateBuilder<ISynonym>();

            using (var iterator = pattern.ToImmutable().GetEnumerator())
            {
                iterator.MoveNext();
                foreach (var word in feed)
                {
                    if (iterator.Current.TryGetSynonym(word, out var s))
                    {
                        matches.Add(s);
                        if (!iterator.MoveNext())
                            break;
                    }
                }
            }
            return new Match(pattern, matches.ToImmutable());
        }
    }
}